'*------------------------------------------------------------------------------*
'* INSTRUCTION MANUAL, AS POSTED TO FORUMS:                                     *
'* Just start typing. There are 37 characters per line, and you move up and     *
'* down through the lines by using the up and down arrow keys. Right now        *
'* there's no way to move left and right through the text with the arrow keys,  *
'* and there's no way to insert text behind other text-- it just overwrites.    *
'*------------------------------------------------------------------------------*

'Written by Eddie Nolan, released under CC-by-sa 3.0
'Legal text available at http://creativecommons.org/licenses/by-sa/3.0/legalcode

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '-----
  'The following constants are for convenience; they are the
  'numerical representations of certain keyboard input keys
  backspace = $C8
  delete = $C9
  space = $20
  enter = $0D
  pageup = $C6
  pagedown = $C7
  upkey = $C2
  downkey = $C3
  '-----
  null = $BC 'This character can be pretty much any non-alphanumerical character,
             'used to represent unprinted characters in some cases
  '-----
  'These characters represent characters in the internal parallax font
  'used for displaying scrollbars, etc.
  uparrow = $04
  downarrow = $05
  tvbackspace = $08
  block = $0E
  line = $7C
  thickline = $7F
  '-----
  clearscreen = $00 'Used with TV.out to clear screen
  '-----
  numberofcharacters = 4000 'Max number of characters that the program can store


OBJ
  KB: "Keyboard"
  TV: "TV_Text-WP"

VAR
  byte x[numberofcharacters]        'The x-value on screen for each character index number
  byte y[numberofcharacters]        'The y-value on screen for each character index number
  byte char[numberofcharacters]     'The character to display for each character index number
  word lineend[120]                 'Save the character index number at which the line ends for each line
  byte linenum                      'Keep track of what line we're on
  byte InputKey       'The key that is input via the keyboard
  byte ydisplay       'Account for hidden lines offscreen
  byte offset         'Number of lines offscreen
  byte tvvalue        'Used in KeyboardPlayback loop
  byte cursorok       'Boolean: Whether or not to go on to a new line at character 37
  word i
  word j
  word k
  word l
  word m              'Miscellaneous variables used in loops

PUB Start
  KB.Startx(26, 27, %0_000_000, %01_01000)
  TV.Start(12)
  WaitCnt(50_000_000 + cnt)
  Repeat                                                'This loop checks to see if the keyboard
    If KB.Present                                       'is present, and tells the user to plug in
        x[0] := 0                                       'the keyboard if not
        y[0] := 0
        i := 0                                                'Start the program at x=0, y=0, char=0
        bytefill(@char, null, 4000)                           'Fill character buffer with 'null' characters
        lineend[0] := 0
        linenum := 0
        offset := 0                                     'Initialize variables
        TV.setxy(0,12)
        TV.str(string("WELCOME            "))           'Display welcome message (line 12 reserved for system messages)
        TV.setxy(0,0)
        DrawScrollbar(1)                                'Draw initial scrollbar
        KeyboardPlayback                                'Begin main loop
    Else
      TV.setxy(0,12)
      TV.str(string("PLUG IN KEYBOARD"))

PRI KeyboardPlayback
  Repeat
    InputKey := KB.GetKey                               'KB.GetKey pauses program until keyboard input recieved
    InputRoutine                                        'This subroutine processes the keyboard input
    If i == 65535
      i := 0  'Prevent bug with delete key
    If x[i] == 37 and cursorok == 0                     'If end of line reached, go on to new line
      lineend[linenum] := i                             'Set the end of current line to current index number
      linenum += 1                                      'Increment line number
      i += 1                                            'Increment index number
      j := i - 1
      ydisplay := y[j] - offset
      TV.setxy(37, ydisplay)
      TV.out(space)                                     'Write space character over cursor
      x[i] := 0
      y[i] := y[j] + 1                                  'Increment y-value
      ydisplay := y[i] - offset
      TV.setxy(x[i],ydisplay)                           'Set x and y
      TV.out(block)                                     'Display cursor
    ElseIf cursorok == 1
      cursorok := 0
    tvvalue := 12 + offset
    If y[i] == tvvalue
      MoveTextDown(y[i])                                'If line 12 reached, move text so that cursor can go down
    If y[i] < offset
      MoveTextUp(y[i])                                  'If cursor goes offscreen, move text so cursor can go up
    TV.setxy(33, 12)                                    'The following code block displays x and y
    TV.dec(x[i])                                        'values in lower-right corner
    TV.out(",")
    TV.dec(y[i])
    'TV.out(",")                                        'Uncomment this code to display variable values for debugging
    'TV.dec(i)
    'TV.out(",")
    'TV.dec(linenum)
    'TV.out(",")
    'TV.dec(offset)
    TV.str(string("  "))
    ydisplay := y[i] - offset
    TV.setxy(x[i], ydisplay)

PRI InputRoutine
  Case InputKey
    backspace:                  'If backspace pressed, go back a space
      j := i - 1
      k := linenum * 38
      k -= 1
      If i <> 0                 'Can't have negative index numbers
        If k <> j               'If not at beginning of line
          TV.out(space)             'Overwrite cursor
          i -= 1                    'De-increment character index
          lineend[linenum] := i
          char[i] := null           'Nullify current character
          linenum := y[i]           'Make sure line number index is correct
          ydisplay := y[i] - offset
          TV.setxy(x[i], ydisplay)  'Set x and y values
          TV.out(block)             'Draw cursor
          TV.out(space)
          DrawScrollbar(linenum)    'Draw scrollbar
        Else                     'If at beginning of line
          TV.out(space)             'Overwrite cursor
          linenum -= 1              'De-increment line number
          i := lineend[linenum]
          ydisplay := y[i] - offset
          TV.setxy(x[i], ydisplay)  'Set x and y values
          TV.out(block)             'Draw cursor
          DrawScrollbar(linenum)    'Draw scrollbar
      If x[i] == 37
        cursorok := 1           'Allow editing of x-value 37

    delete:                     'If delete key pressed, go back a space (identical to backspace routine)
      j := i - 1
      k := linenum * 38
      k -= 1
      If i <> 0                 'Can't have negative index numbers
        If k <> j               'If not at beginning of line
          TV.out(space)             'Overwrite cursor
          i -= 1                    'De-increment character index
          lineend[linenum] := i
          char[i] := null           'Nullify current character
          linenum := y[i]           'Make sure line number index is correct
          ydisplay := y[i] - offset
          TV.setxy(x[i], ydisplay)  'Set x and y values
          TV.out(block)             'Draw cursor
          TV.out(space)
          DrawScrollbar(linenum)    'Draw scrollbar
        Else                     'If at beginning of line
          TV.out(space)             'Overwrite cursor
          linenum -= 1              'De-increment line number
          i := lineend[linenum]
          ydisplay := y[i] - offset
          TV.setxy(x[i], ydisplay)  'Set x and y values
          TV.out(block)             'Draw cursor
          DrawScrollbar(linenum)    'Draw scrollbar
      If x[i] == 37
        cursorok := 1           'Allow editing of x-value 37

    enter:                      'If enter key pressed, go down a line
      TV.out(space)             'Overwrite cursor
      lineend[linenum] := i     'Set line-end index to current character index value
      linenum += 1              'Increment line number
      j := i
      i := (linenum * 38)       'Character index value for first character of every line
                                'is equal to line number * 38; this prevents overwriting
                                'of other characters
      x[i] := 0
      y[i] := y[j] + 1
      ydisplay := y[i] - offset
      TV.setxy(x[i],ydisplay)
      TV.out(block)             'Set x and y values and draw cursor

    upkey:
      MoveCursorUp
      DrawScrollbar(linenum)

    downkey:
      MoveCursorDown
      DrawScrollbar(linenum)

    OTHER:                      'If normal key pressed, print character
      ydisplay := y[i] - offset
      TV.setxy(x[i], ydisplay)  'Set x and y
      char[i] := InputKey       'Set character for current index value to input character
      TV.out(char[i])           'Display character
      TV.out(block)             'Display cursor
      j := i
      i += 1
      x[i] := x[j] + 1
      y[i] := y[j]
      lineend[linenum] := i     'Increment character index value and line-end value
      DrawScrollbar(linenum)    'Draw scrollbar

PRI MoveCursorUp
  j := linenum - 1
  k := lineend[j]
  If y[i] > 0                   'Prevent moving cursor up on line zero
    linenum := j                'De-increment line number
    i := k                      'Set character index value to line-end value of previous line number
    TV.out(space)               'Overwrite cursor
    i := lineend[linenum]       'Set cursor to line-end index of previous line
    ydisplay := y[i] - offset
    TV.setxy(x[i], ydisplay)
    TV.out(block)               'Set x and y values and print cursor
    If x[i] == 37               'Allow editing of 37th character
      cursorok := 1
  Else                          'Special case with 0th line
    ydisplay := y[i] - offset
    TV.setxy(x[i], ydisplay)
    If x[i] == 37
      cursorok := 1

PRI MoveCursorDown
  j := y[i] + 1
  k := lineend[j]
  If k > 0                      'If the next line exists
    i := k                      'Character index is set to end of next line
    linenum := y[i]             'Keep track of line number
    TV.out(space)               'Overwrite cursor
    ydisplay := y[i] - offset
    TV.setxy(x[i], ydisplay)
    TV.out(block)               'Set x and y and draw cursor
    If x[i] == 37
      cursorok := 1             'Allow editing of 37th character

PRI MoveTextDown (linenumber)
  TV.out(clearscreen)
  linenum := linenumber         'Set current line number to variable passed into subroutine
  offset += 1                   'Increase line offset
  k := offset - 1
  k := lineend[k]
  k += 1
  Repeat m from k to i          'This loop writes all the characters to the screen
    ydisplay := y[m] - offset
    TV.setxy(x[m], ydisplay)
    If char[m] <> null          'Ignore character if 'null'
      TV.out(char[m])
  ydisplay := y[i] - offset
  TV.setxy(x[i], ydisplay)
  TV.out(block)
  DrawScrollbar(linenum)        'Set x and y and draw cursor and scrollbar

PRI MoveTextUp (linenumber)
  TV.out(clearscreen)
  offset -= 1                   'Decrease line offset
  linenum:= linenumber
  j := offset * 38
  l := offset + 11
  repeat m from j to lineend[l] 'Write all characters to screen
    ydisplay := y[m] - offset
    TV.setxy(x[m], ydisplay)
    If char[m] <> null          'Ignore if 'null'
      TV.out(char[m])
  ydisplay := y[i] - offset
  TV.setxy(x[i], ydisplay)
  TV.out(block)
  DrawScrollbar(linenum)        'Set x and y and draw cursor and scrollbar

PRI DrawScrollbar (linenumber)
  TV.setxy(39,0)
  TV.out(uparrow)
  TV.setxy(39,12)
  TV.out(downarrow)
  repeat m from 1 to 11
    TV.setxy(39,m)
    TV.out(line)
  m := 0
  k := 0
  l := 0
  repeat until m == 1           'This loop finds the current line number, multiplies by 11
    If lineend[k] == 0          'and divides by the total number of lines to decide where to
      l := k + 1                'write the scrollbar character
      If lineend[l] == 0
        m := linenum * 11
        l := m / k
        l += 1
        TV.setxy(39,l)
        TV.out(thickline)
        Return
    k += 1
  ydisplay := y[i] - offset
  TV.setxy(x[i], ydisplay)

